/**
 * AccountDetails wrapper class.
 * @param   data Example : {"credits":{"total":4652,"subscription":4647,"payg":5},"api":{"free_calls":50,"sizes":["regular","medium","hd","4k"]}}
 * @returns {{getMaxSize: (function(): string), getPayG: (function(): *), getFreeCalls: (function(): *), getTotalCredits: (function(): (PaymentItem | number)), getSubscription: (function(): *), getSizes: (function(): (string | DOMTokenList))}}
 * @constructor
 */
var AccountDetails = function(data) {

    if (typeof(data) == 'undefined' || typeof(data.attributes) == 'undefined') {
        throw new AccountDetailsError('Data object is undefined');
    }

    var account = data.attributes;

    return {
        /**
         * Get total credits balance.
         * @returns {PaymentItem | number}
         */
        getTotalCredits : function() {
            if (typeof(account.credits.total) == 'undefined') {
                throw new AccountDetailsError('Account.credits.total is undefined');
            }
            return account.credits.total;
        },

        /**
         * Get subscription balance.
         * @returns {*}
         */
        getSubscription : function() {
            if (typeof(account.credits.subscription) == 'undefined') {
                throw new AccountDetailsError('Account.credits.subscription is undefined');
            }
            return account.credits.subscription;
        },

        /**
         * Get pay-as-you-go balance.
         * @returns {*}
         */
        getPayG : function() {
            if (typeof(account.credits.payg) == 'undefined') {
                throw new AccountDetailsError('Account.credits.payg is undefined');
            }
            return account.credits.payg;
        },

        /**
         * Get subscription sizes.
         * @returns {string | DOMTokenList}
         */
        getSizes : function() {
            if (typeof(account.api.sizes) == 'undefined') {
                throw new AccountDetailsError('Account.api.sizes is undefined');
            }
            return account.api.sizes;
        },

        /**
         * Get max size available for account.
         * @returns {string}
         */
        getMaxSize : function() {
            /*
             * Kaleido decided to allow all accounts to use the 4k max size so
             * the max size check is no longer needed. Rather than refactor all
             * of the code, for now I am simply hard-coding the max size to the
             * highest possible value. This is being done  to minimize the cost,
             * level of effort, and testing needed  to implement the change.
             * This change was implemented with only 72 hours before launch.
             * Making this sweeping of a refactor would have raised too much
             * risk and required too much testing to safely do so close to launch.
             * This is understandably a kludge but in this case I think it is
             * the safest route to go until there is sufficient time to
             * safely make the code changes needed to remove this feature.
             */
            return '4k';

            // if (typeof(account.api.sizes.length) == 'undefined') {
            //     throw new AccountDetailsError('Account.credits.sizes.length is undefined');
            // }
            // return account.api.sizes[account.api.sizes.length-1];
        },

        /**
         * Get free API calls.
         * @returns {*}
         */
        getFreeCalls : function() {
            if (typeof(account.api.free_calls) == 'undefined') {
                throw new AccountDetailsError('Account.api.free_calls is undefined');
            }
            return account.api.free_calls;
        },

        /**
         * Determine if account has available credits.
         * @returns {boolean}
         */
        hasCredits : function() {
            return Number(account.credits.total) > 0;
        },

        /**
         * Determine if available balance is enough for required credits.
         * @param needed
         * @returns {boolean}
         */
        hasEnoughCredits : function(needed) {
            return Number(account.credits.total) >= needed;
        },

        /**
         * Determine if size is allowed.
         * @param theSize
         */
        isAllowedSize : function(theSize) {

            var values = {
                xregular : 100,
                xmedium  : 200,
                xhd      : 300,
                x4k      : 400
            };

            var maxSize = account.api.sizes[account.api.sizes.length-1];

            if (theSize <= 100) {
                size = "regular";
            }
            else if (theSize <= 200) {
                size = "medium";
            }
            else if (theSize <= 300) {
                size = "hd";
            }
            else if (theSize <= 400) {
                size = "4k";
            }
            else {
                // The image is too large.
            }

            // regular, medium, hd, 4k
            //@TODO: Determine if size is in available range.
        },

        /**
         * Get the data object.
         * @returns {ActiveX.ISchemaItemCollection | boolean | NamedNodeMap | ActiveX.IXMLDOMNamedNodeMap | *}
         */
        getData : function() {
            return data.attributes;
        },

        /**
         * Get data JSON as string.
         * @returns {string}
         */
        toString : function() {
            return JSON.stringify(data.attributes);
        }
    }
};

/**
 * Custom account details error.
 * @param message
 * @param file
 * @param line
 * @constructor
 */
var AccountDetailsError = function(message, file, line) {
    return {
        message : message || "Account Details error",
        file    : "",
        line    : ""
    }
}
AccountDetailsError.prototype = Error.prototype;
