/**
 * Global config object.
 * @type {{}}
 */
window.Config = window.Config || {};

/**
 * A global object of constants.
 * @type {{EXTENSION_ID: string}}
 */
window.Config.extensionId = csInterface.getExtensionID();

/**
 * Global counter.
 * @type {number}
 */
window.counter = window.counter || 0;

/**
 * Image resolutions constants.
 * @type {{"4k": number, medium: number, hd: number, regular: number}}
 */
// window.resolutions = window.resolutions || {
//     "regular" : 250000,
//     "medium"  : 1500000,
//     "hd"      : 4000000,
//     "4k"      : 10000000,
// };
/**
 * @see AccountDetails.js::getMaxSize() for detailed explanation of these values.
 * @type {*|{"4k": number, medium: number, hd: number, regular: number}}
 */
window.resolutions = window.resolutions || {
    "regular" : 10000000,
    "medium"  : 10000000,
    "hd"      : 10000000,
    "4k"      : 10000000,
};

/**
 * Persistance object.
 * @param inOn
 * @constructor
 */
var Persistent = function(inOn) {

    var event,
        persistance;

    // Make the persistance false by default.
    persistance = "com.adobe.PhotoshopUnPersistent";

    // If inOn is true, change the value.
    // (Doing it this way means less code to be executed).
    if (inOn) {
        var event = new CSEvent("com.adobe.PhotoshopPersistent", "APPLICATION");    //Photoshop
        persistance = "com.adobe.PhotoshopUnPersistent";
    }

    event = new CSEvent(persistance, "APPLICATION");
    event.extensionId = "bg.remove.panel";
    csInterface.dispatchEvent(event);
}

/**
 * Creates a string ID
 * @returns {string}
 */
function getId() {
    return String(counter++);
}

/**
 * Enable the main button.
 */
function enableBtn() {
    $("#remove-bg-btn").text('Remove Background').attr("disabled", false);
}

/**
 * Disables the main button.
 */
function disableBtn() {
    $("#remove-bg-btn").attr("disabled", true);
}

/**
 * Test if a status message is an image resolution error.
 * @param status
 * @returns {boolean}
 */
function isImageResolutionError(status) {
    return status.toLowerCase() == 'max_res_exceeded';
}

/**
 * Test if a status message is a resize image error.
 * @param   {string}    status
 * @returns {boolean}
 */
function isResizeImageError(status) {
    return status.toLowerCase() == 'resize_image_error';
}

/**
 * Deletes the temporary file used by the extension.
 * @param {filename}    The temporary file used by the API.
 */
function deleteRemoveBgFile(filename) {
    try {
        var result = window.cep.fs.deleteFile(filename);
        if (result.err !== 0) {
            throw new DeleteFileError("DeleteFileError : " + result.err);
        }
    }
    catch(e) {
        info(e);
    }
}

/**
 * Read a file.
 * @param theFilePath
 * @returns {*}
 */
function readFileData(theFilePath) {
    try {
        // window.cep.fs.readFile(path, cep.encoding.Base64);
        var result = window.cep.fs.readFile(theFilePath);
        if (result.err !== 0) {
            throw new Error("ReadFileError : " + result.err);
        }
        return result.data;
    }
    catch(e) {
        info(new ReadFileError(e.message, e.stack));
        throw new ReadFileError(e.message, e.stack);
    }
}

/**
 * Build the custom User Agent string.
 * @param extensionId
 * @returns {*}
 */
function getUserAgent(extensionId) {

    var theAppInfo,
        theExtInfo;

    try {
        theAppInfo = getApplicationInfo();
    }
    catch(e) {
        info(ApplicationInfoError(e));
    }

    try {
        theExtInfo = getExtensionInfo(extensionId);
    }
    catch(e) {
        info(ExtensionInfoError(e));
    }

    var tokens = {
        ID : get(theAppInfo, 'ID', '??'),

    }

    return _t(
        '{ID} {PS} Plugin {EXT}/CEP {CEP} {OS}',
        {
            ID  : get(theAppInfo, 'ID', '??'),
            PS  : get(theAppInfo, 'Version', '??'),
            CEP : get(theAppInfo, 'CEP_Runtime', '??'),
            EXT : get(theExtInfo, 'version', '??'),
            OS  : csInterface.getOSInformation()
        }
    )
}

/**
 * Get the installData dump.
 * @type {string}
 */
function getInstallData() {
    try {
        return readFileData(
            csInterface.dumpInstallationInfo()
        );
    }
    catch(e) {
        throw new InstallInfoError(e.message, e.stack);
    }
}

/**
 * Parse the Extension manifest.
 * @returns {{Version: string}}
 */
function getExtensionInfo() {

    var extensionData = {
        version : ''
    };

    try {

        var ext,
            xmlString,
            theManifest;

        ext = getExtension(csInterface.getExtensionID());

        if (get(ext, 'basePath', false)) {

            xmlString   = readFileData( path(ext.basePath, 'CSXS/manifest.xml'));
            theManifest = $.parseXML(xmlString);
            var $ext    = $('Extension[Id="' + csInterface.getExtensionID() + '"]', theManifest).eq(0);

            extensionData.version = $ext.attr('Version');
        }
    }
    catch(e) {
        info(new ExtensionInfoError(e.message, e.stack));
        throw new ExtensionInfoError(e.message, e.stack);
    }

    return extensionData;
}

/**
 * Get the App info from Extension Installation Info dump.
 * @returns {{Locale: string, CEP_Runtime: string, Version: string, ID: string}}
 */
function getApplicationInfo() {

    var appInfo = {
        ID          : '',
        Version     : '',
        Locale      : '',
        CEP_Runtime : ''
    };

    var patterns = {
        'ID'          : new RegExp(/ID:(.*)/i),
        'Version'     : new RegExp(/Version:(.*)/i),
        'Locale'      : new RegExp(/Locale:(.*)/i),
        'CEP_Runtime' : new RegExp(/CEP Runtime Version:(.*)/i)
    };

    var theDump;

    try {
        theDump = getInstallData();

        for (key in patterns) {

            var found = theDump.match( patterns[key] );

            if (typeof (found.length) != 'undefined' && found.length > 1) {
                appInfo[key] = found[1];
            }
        }
    }
    catch(e) {
        info(new ApplicationInfoError(e.message, e.stack));
        throw new ApplicationInfoError(e.message, e.stack);
    }
    return appInfo;
}

/**
 * Determine if application is currently connected to the internet.
 * @returns {*}
 */
function isAppOnline() {
    return get(csInterface.getNetworkPreferences(), 'isAppOnline', false);
}

/**
 * Set UI state to 'uploading'.
 */
function setUploadingState() {
    disableBtn();
    $('body').removeClass('uploading');
    $('body').addClass('uploading');
}

/**
 * Set UI state to ready.
 */
function setReadyState() {
    enableBtn();
    $('body').removeClass('uploading');
}


/**
 * For dev only. Test the progress bar.
 */
function startProgressBar() {

    var inProgress = false,
        $progress  = $("#progress");

    if (! inProgress) {

        updateProgress(0);

        var timerId = setInterval(function() {

            inProgress = true;
            var value  = parseInt( $progress.attr('value') );

            updateProgress(value + 1);

            if ( value >= 100) {
                clearInterval(timerId);
                inProgress = false;
            }
        }, 200);
    }
}

/**
 * Update the progress bar values.
 * @param value
 * @param max
 */
function _updateProgress(value) {
    $("#progress").attr('value', value);
}

function updateProgressValue(value) {
    var $progress  = $("#progress");
    $('#indicator').css({
        width: (($progress.data('value') / $progress.data('max')) * 100) + "%"
    });
}

/**
 * Update the progress bar from the xhr.progress event.
 * @param evt
 */
function updateProgressBar(loaded, total, direction) {

    var $progress = $("#progress");
    var progress = (loaded / total) * 100;

    info(direction + ' progress : ' + progress + '%');

    $("#progress").attr('value', progress);
    // updateProgressValue(progress);
}

/**
 * Common interface for handling API status code returns.
 * @type {{"400": statusCodeHandler.400, "402": statusCodeHandler.402, "403": statusCodeHandler.403, "404": statusCodeHandler.404, "429": statusCodeHandler.429}}
 */
function statusCodeMessage(theCode) {
    switch(theCode) {
        case 400 :
            return 'Error: Invalid parameters or input file unprocessable (no credits charged)';
        case 402 :
            return 'Error: Insufficient credits (no credits charged)';
        case 403 :
            return 'Error: Authentication failed or API not available for this account (no credits charged)';
        case 404 :
            return 'Page not found';
        case 429 :
            return 'Error: Rate limit exceeded (no credits charged)';
        default:
            return 'An unknown error occcured';
    }
}

/**
 * Callback for showing error status code messages and resetting UI.
 * @param statusCode
 */
function statusCodeHandler(statusCode) {
    var error = statusCodeMessage(statusCode);
    info(error);
    csInterface.evalScript('host.logger.error("' + error + '")', function(result) {
        csInterface.evalScript('host.alert("' + error + '")');
    });
    removeProgressBar();
}

/**
 * Replace tokens in a string with key => value paired vars.
 * @param theText
 * @param theVars
 * @returns {*}
 * @private
 */
function _t(theText, theVars) {
    for (token in theVars) {
        theText = theText.replace(
            new RegExp("{" + token + "}","g"),
            theVars[token]
        );
    }
    return theText;
}

/**
 * Executes the main API call.
 */
function removeBackground(maxSize) {

    var maxRes =  resolutions[maxSize];

    csInterface.evalScript("verifyFileSize(" + maxRes + ")", function(status) {
        try {

            if (! isJSON(status)) {
                var statusError = new Error(_t(
                    'Improperly formatted JSON response in functions.js:409 - {s}',
                    {s: status}
                ));
                error(statusError);
                throw statusError;
            }

            status = JSON.parse(status);

            if (status.status != "ok") {
                var removeBackgroundError = new RemoveBackgroundError(_t(
                    "removeBackground() encountered an error {err}",
                    {err: status.error}
                ));
                error(removeBackgroundError);
                throw removeBackgroundError;
            }

            path = status.path;

            info( 'Got path : {s}', {s: path});

            var readResult = window.cep.fs.readFile(path, cep.encoding.Base64);

            if (readResult.err != 0) {
                // error(new ReadFileError("Failed to read file: " + readResult.err));
                error(new ReadFileError("Failed to read file: " + readResult.err));
            }

            info("Read file: " + readResult.data.length + " bytes (b64)");

            var formData = new FormData();
                formData.append(
                    "image_file",
                    b64toBlob(readResult.data, 'image/jpeg'),
                    "original.jpg"
                );
                formData.append("size", "auto");

            $.ajax({

                // Request settings.

                type        : "POST",
                url         : Config.API_REMOVE_BG,
                data        : formData,
                cache       : false,
                contentType : false,
                processData : false,

                // Request headers.

                headers     : {
                    'X-Api-Key'    : getApiKey(),
                    'Accept'       : 'application/json',
                    'X-User-Agent' : Config.USER_AGENT
                },

                // Handle returned status codes

                statusCode: {
                    400 : function() { statusCodeHandler(400); },
                    402 : function() { statusCodeHandler(402); },
                    403 : function() { statusCodeHandler(403); },
                    404 : function() { statusCodeHandler(404); },
                    429 : function() { statusCodeHandler(429); }
                },

                // Add the progress bar updater

                xhr: function() {

                    var xhr = new window.XMLHttpRequest();

                    // set the onprogress event handler

                    xhr.upload.addEventListener("progress", function(evt) {
                        $("#remove-bg-btn").text('Uploading ...');
                        showProgressBar();
                        if (evt.lengthComputable) {
                            updateProgressBar(evt.loaded, evt.total, 'Upload');
                        }
                    }, false);

                    xhr.addEventListener("progress", function(evt) {
                        $("#remove-bg-btn").text('Downloading ...');
                        if (evt.lengthComputable) {
                            updateProgressBar(evt.loaded, evt.total, 'Download');
                        }
                    }, false);

                    return xhr;

                },

                // Handle success message from API.

                success: function (data, textStatus, xhr) {

                    var filename,
                        saveResult,
                        xCreditsCharged,
                        xWidth,
                        xHeight,
                        megapixels;

                    // X-Credits-Charged

                    xCreditsCharged = xhr.getResponseHeader('X-Credits-Charged');
                    xWidth          = xhr.getResponseHeader('x-width');
                    xHeight         = xhr.getResponseHeader('x-height');

                    info('X-Credits-Charged : {s}', {s: xCreditsCharged});
                    info('x-width : {d}',           {d: xWidth});
                    info('x-height : {d}',          {d: xHeight});
                    info(xhr);

                    // Get the result file path.

                    filename = csInterface.getSystemPath(SystemPath.USER_DATA) + "/removebg-result.png";

                    // Get the result of saving the file sent back from the API.

                    saveResult = window.cep.fs.writeFile(
                        filename,
                        data.data.result_b64,
                        cep.encoding.Base64
                    );

                    info("Saving to " + filename + " ...");

                    // Handle result of saving the file.

                    if (saveResult.err === 0) {
                        info("Result file saved: " + filename);
                        csInterface.evalScript("addLayerFromFile('" + filename + "')", function(status) {
                            deleteRemoveBgFile(filename);
                        });

                        // Display the number of credits charged for removing the background.

                        $(".credits-charged").text(xCreditsCharged + " credits");

                        var $balance = $(".balance");

                        var balance = Number(parseInt($balance.text()));
                            balance = Number(balance - xCreditsCharged);

                        $balance.text('');
                        $balance.text(balance + ' credits');

                        if (isNumber(xHeight) && isNumber(xWidth)) {
                            $(".image-size").text(
                                calculateMegaPixels(Number(xWidth), Number(xHeight)) + ' MP'
                            );
                        }

                        showTemporarily('.credits-charged-row', 30);
                        showTemporarily('.image-size-row', 30);
                    }
                    else {
                        var saveResultError = new SaveResultError(
                            _t('Failed to save result: {e}', {e: saveResult.err})
                        );
                        error(saveResultError);
                        throw saveResultError;
                    }

                    setUserInterfactStatus(InterfaceStatus.READY);
                },

                // Handle API call error.

                error: function(e) {
                    if (e.responseJSON && e.responseJSON.errors) {

                        var msg = [];
                        for (var i = 0; i < e.responseJSON.errors.length; i++) {
                            msg.push(e.responseJSON.errors[i].title);
                        }

                        csInterface.evalScript("host.alert('" + msg.join("\n") + "')");
                        error(msg);
                    }
                    else {
                        error(
                            new RemoveBackgroundError(
                                "Error: {e}".replace('{e}', JSON.stringify(e) )
                            )
                        );
                    }
                    csInterface.evalScript("cancelRemoveBg()");
                    setUserInterfactStatus(InterfaceStatus.READY);
                },

                // Display the progress bar before initiating the request.

                beforeSend: function( xhr ) {
                    setUserInterfactStatus(InterfaceStatus.UPLOADING);
                }
            });
        }
        catch(e) {
            info(e);
            setUserInterfactStatus(InterfaceStatus.READY);
        }
    });
}

/**
 * UI Statuses.
 * @type {{DISABLED: string, READY: string, UPLOADING: string}}
 */
var InterfaceStatus = {
    UPLOADING : 'uploading',
    DISABLED  : 'disabled',
    READY     : 'ready'
};

/**
 * Set the user interfact status.
 * @param theStatus
 */
function setUserInterfactStatus(theStatus) {
    if (theStatus == InterfaceStatus.UPLOADING) {
        $('body').removeClass('uploading disabled ready');
        $('body').addClass('uploading');
        disableBtn();
        showProgressBar();
    }
    else if (theStatus == InterfaceStatus.DISABLED) {
        $('body').removeClass('uploading disabled ready');
        $('body').addClass('disabled');
        removeProgressBar();
        disableBtn();
    }
    else if (theStatus == InterfaceStatus.READY) {
        $('body').removeClass('uploading disabled ready');
        $('body').addClass('ready');

        removeProgressBar();
        enableBtn();
    }
    else {
        throw new Error('Not a valid InterfaceStatus');
    }
}

/**
 * Show an element for a limited time.
 * @param selector
 * @param duration
 */
function showTemporarily(selector, duration) {
    var $item = $(selector);
    $item.fadeIn(200, function() {
        setTimeout(function() {
            $item.fadeOut(200);
        }, duration * 1000);
    });
}

function getAccountDetails(refresh) {
    // https://api.remove.bg/v1.0/account

    var account,
        details;

    try {
        $.ajax({

            // Request settings.

            type        : "GET",
            url         : 'https://api.remove.bg/v1.0/account',
            cache       : false,
            async       : false,
            contentType : false,
            processData : false,

            // Request headers.

            headers     : {
                'X-Api-Key' : getApiKey(),
                'Accept'    : 'application/json'
            },

            // Handle returned status codes

            statusCode: {
                400 : function() { statusCodeHandler(400); },
                402 : function() { statusCodeHandler(402); },
                403 : function() { statusCodeHandler(403); },
                404 : function() { statusCodeHandler(404); },
                429 : function() { statusCodeHandler(429); }
            },

            // Handle success message from API.

            success: function (data, textStatus, xhr) {

                var account = new AccountDetails(data.data);

                Config.ACCOUNT = account;

                info('Total credits : ' + account.getTotalCredits());
                info('Subscription  : ' + account.getSubscription());
                info('Max size      : ' + account.getMaxSize());
                info('Get Sizes     : ' + account.getSizes());

                $("#account-data").val(account.toString());

                info($("#account-data").val());

                var credits = account.getTotalCredits();
                $(".balance").text(credits + " credits");
                $("#balance-header").text('You have ' + credits + ' available');
                $(".max-size").text(account.getMaxSize());
            },

            // Handle API call error.

            error: function(e) {
                if (e.responseJSON && e.responseJSON.errors) {
                    var msg = [];
                    for (var i = 0; i < e.responseJSON.errors.length; i++) {
                        msg.push(e.responseJSON.errors[i].title);
                    }
                    msg = msg.join("\n");
                    csInterface.evalScript(
                        "host.logger.error('{s}')".replace('{s}', msg)
                    );
                    error(msg);
                }
                else {
                    error(new Error("Error: " + e));
                }
            }
        });
    }
    catch(e) {
        info(e);
        enableBtn();
    }
}

/**
 * Get the current extension version.
 * @returns {string}
 */
function getExtensionVersion() {
    var extensionInfo = getExtensionInfo();
    var version = '0.0.0';
    if (typeof extensionInfo == 'object' && typeof extensionInfo.version != 'undefined') {
        version = extensionInfo.version;
    }
    return version;
}

/**
 * API Call to make sure the current extension is still supported.
 *
 * API Endpoint : GET https://api.remove.bg/v1.0/apps/photoshop?v=1.0.0
 *
 * OK:
 *     - All is good.
 *     {"status": "ok"}
 *
 * WARNING:
 *     - Show a dismissable warning. Extension can still run for now.
 *     {
 *         "status": "warning",
 *         "message": "You are using an outdated version. This version will stop working in 3 weeks. Please update at www.remove.bg/photoshop"
 *     }
 *
 * ERROR:
 *     - Show a non-dismissable warning, that completely disables the extension underneath it
 *     {
 *         "status": "error",
 *         "message": "This plugin is no longer supported. Please update at www.remove.bg/photoshop"
 *     }
 * @returns {*}
 */
function verifyExtensionVersion() {
    try {
        var xhr = $.ajax({
            type        : "GET",
            url         : 'https://api.remove.bg/v1.0/apps/photoshop?v=' + getExtensionVersion(),
            cache       : false,
            async       : false,
            contentType : false,
            processData : false,

            // Request headers.

            headers     : {
                'Accept'    : 'application/json'
            },

            success: function (data, textStatus, xhr) {
                console.log('Version Data', data);
            },

            // Handle API call error.

            error: function(e) {
                if (e.responseJSON && e.responseJSON.errors) {
                    var msg = [];
                    for (var i = 0; i < e.responseJSON.errors.length; i++) {
                        msg.push(e.responseJSON.errors[i].title);
                    }
                    msg = msg.join("\n");
                    csInterface.evalScript(
                        "host.logger.error('{s}')".replace('{s}', msg)
                    );
                    error(msg);
                }
                else {
                    error(new Error("Error: " + e));
                }
            }
        });
        return xhr.responseJSON;
    }
    catch(e) {
        error(e);
        return {status: 'error', message: e.message};
    }
}

/**
 * Loads external JSX files.
 */
function loadDependencies() {
    var ext = getExtension(csInterface.getExtensionID());
    var files = [
        'host/prototypes.js',
        'host/JSON.jsx',
        'host/Logger.jsx',
        'host/Helpers.jsx',
        'host/SelectionDimensions.jsx',
        'host/Utils.jsx',
        'host/SettingsDialog.jsx',
        'host/ErrorClasses.jsx',
        'host/globals.js'
    ];
    for (var i=0; i<files.length; i++) {
        var theFile = ext.basePath +  files[i];
        csInterface.evalScript("include('" + theFile + "')", function(result) {
            info(result);
        });
    }
}

/**
 * tests to see if the API key is set/saved.
 * @returns {boolean}
 */
function hasValidApiKey() {
    try {
        var apiKey = getApiKey();
        info('apiKey : ' + apiKey);
        info('apiKey isDefined : ' + isDefined(apiKey) );
        info('apiKey isString : ' + isString(apiKey) );
        info('apiKey isEmpty : ' + isEmpty(apiKey) );
        if (! isDefined(apiKey)) return false;
        if (! isString(apiKey))  return false;
        if (isEmpty(apiKey))     return false;

        return true;
    }
    catch(e) {
        info(new Error('hasValidApiKey error (functions.js:672) ' + e));
        return false;
    }
    // return isNonEmptyString(apiKey);
}

/**
 * Adds "Configure API key" message header to palette.
 */
function showConfigureApiKeyMessage() {
    removeTopSectionMessage();
    $("#top-section").append(
        '<h2 id="configure-api-key">Please configure an API key in the plugin settings</h2>'
    );
}

/**
 * Removes messages from palette header.
 */
function removeTopSectionMessage() {
    $('.message', '#top-section').remove();
}

/**
 * Adds "Configure API key" message header to palette.
 */
function showExtensionVersionMessage(apiResponse) {

    removeBlockingOverlay();

    // (apiResponse.status == 'warning' ? '<span class="icomatic icon">cancel</span>' : '' ) +

    $('#logo').before(_t(
        '<div id="{id}" class="message {status}">' +
        '<p>{message}<br>{button}</p></div>', {
            id      : 'ui-blocking-overlay',
            status  : apiResponse.status,
            message : apiResponse.message.replace(
                'www.remove.bg/photoshop',
                '<a href="#" onclick="openExtensionPage()">www.remove.bg/photoshop</a>'
            ).replace('supported', '<br>supported'),
            button  : overlayButton(apiResponse.status)
        }
    ));

    if (apiResponse.status == 'warning') {
        $('#version-check-btn').on('click', function(e) {
            $('#ui-blocking-overlay').fadeOut('slow', function() {
                $(this).remove();
                setUserInterfactStatus(InterfaceStatus.READY);
            });
        });
    }
    else if (apiResponse.status == 'error') {
        $('#version-check-btn').on('click', openExtensionPage);
    }
}

/**
 * Get HTML for version warning/error overlay button.
 * @param   {string}    status  The API response status
 * @returns {*}
 */
function overlayButton(status) {

    if (['warning', 'error'].indexOf(status) == -1) return '';

    var theText = 'Okay';

    if (status == 'error') {
        theText = 'Update Now';
    }

    return _t(
        '<button id="version-check-btn" class="topcoat-button--large--cta">\n' +
        '    <span class="label">' + theText + '</span>\n' +
        '</button>'
    )
}

/**
 * Open the Extension product page.
 */
function openExtensionPage() {
    csInterface.openURLInDefaultBrowser(
        'https://www.remove.bg/photoshop'
    );
}

/**
 * Removes the extension version warning overlay.
 */
function removeBlockingOverlay() {
    $('#ui-blocking-overlay').remove();
}

/**
 * Adds the progress bar to the palette.
 */
function showProgressBar() {
    removeProgressBar();
    $("#top-section").append(
        '<progress id="progress" max="100" value="0"></progress>'
    );
}


/**
 * Removes the progress bar from the palette.
 */
function removeProgressBar() {
    $("#progress").remove();
}

/**
 * Calculate megapixels from width and height of an image.
 * @param   {number}    width
 * @param   {number}    height
 * @returns {number}
 */
function calculateMegaPixels(width, height) {
     var mp = (width * height) / 1000000;
     return Math.round((mp + 0.00001) * 100) / 100 ;
}

/**
 * Scale an image by megapixels.
 * @param   {number}    width   The original width
 * @param   {number}    height  The original height
 * @param   {number}    target  The target size in megapixels.
 * @returns {number[]}
 */
function scaleByMegapixels(width, height, target) {
    var mp    = calculateMegaPixels(width, height);
    var scale = Math.min(mp, target) / Math.max(mp, target);

    return [
        (width * width) * scale,
        (height * height) * scale
    ]
}

/**
 * Write message to the in-palette log.
 * @deprecated
 * @param msg
 */
function log(msg) {

    var $log = $("#log");

    $log.val( $log.val() + "\n" + msg );
}

/**
 * Get an extension object. Defaults to current extension.
 * @param extensionId
 * @returns {*}
 */
function getExtension( extensionId ) {
    var extension,
        extensions;

    if (typeof(extensionId) == 'undefined') {
        extensionId = csInterface.getExtensionId();
    }

    extensions = csInterface.getExtensions( [extensionId] );

    if ( extensions.length == 1 ) {

        extension = extensions[0];
        extension.basePath = slash( extension.basePath );
        info(extension.basePath);
    }

    return extension;
}

/**
 * Interface to Host.showSettingsDialog()
 */
function showSettingsDialog() {

    // Listen for the 'Profile' button click.

    csInterface.removeEventListener(Events.OPEN_URL, eventHandlerOpenInBrowser);
    csInterface.addEventListener(Events.OPEN_URL, eventHandlerOpenInBrowser);

    // Listen for the 'Save API Key' button click.

    csInterface.removeEventListener(Events.KEY_SAVED, reloadExtension);
    csInterface.addEventListener(Events.KEY_SAVED, reloadExtension);

    // Show the settings dialog.

    csInterface.evalScript( 'host.showSettingsDialog()', function(result) {
        info('host.showSettingsDialog returned')
        info(result);
    });
}

/**
 * Flyout menu builder.
 */
function initFlyoutMenu() {
    console.log(Config);
    console.log('Is Debug ?', isDebug());
    var menuItems = [
        {id : 'settings', label : 'Settings', enabled : true, checked : false}
    ];

    if (isDebug() === true) {
        menuItems = [
            {id : 'settings', label : 'Settings', enabled : true, checked : false},
            {id : 'reload',   label : 'Reload Extension', enabled : true, checked : false}
        ];
    }
    new FlyoutMenu({
        items : menuItems,
        onClick : function(event) {
            switch (event.data.menuId) {
                case "settings":
                    showSettingsDialog();
                    break;
                case 'reload':
                    reloadExtension();
                    break;
                default:
                    break;
            }
            info(event.data.menuName + " clicked!");
        }
    }).build();
}

/**
 * Are we in debug mode?
 * @returns {boolean}
 */
function isDebug() {
    return Config.DEBUG === true;
}

/**
 * DEV ONLY
 * Reloads the extension front-end.
 */
function reloadExtension() {
    try {
        window.cep.process.removeAllListeners();
        window.location.href = "index.html";
    }
    catch (e) {
        window.location.href = "index.html";
    }
}

/**
 * Open url in browser.
 * @param event
 */
function eventHandlerOpenInBrowser(event) {
    info(event.data);
    csInterface.openURLInDefaultBrowser(event.data.url);
}
