/**
 * Create symbols for classes in global scope.
 */
var host,
    Config,
    Utils,
    Logger,
    logger,
    SettingsDialog;

var originalDocument  = null,
    originalSelection = null,
    tempDoc           = null,
    hasSelection      = false,
    wasResized        = false,
    selectionChannel  = null;


var Host = function(Config) {

    var ExternalObjectIsLoadedLib = false;
    var _logger;

    _logger = new Logger(Config.EXTENSION_ID, Config.LOG_FOLDER);
    _logger.clear();

    logger = _logger;

    /**
     * Get the API key
     * @returns {*}
     * @private
     */
    function _getApiKey() {

        var settings;

        _logger.info(JSON.stringify(Config));
        _logger.info('Config.SETTINGS_FILE : ' + Config.SETTINGS_FILE);
        _logger.info('Settings file exists : ' + (new File(Config.SETTINGS_FILE).exists));

        try {
            settings = Utils.read_json_file(Config.SETTINGS_FILE);
            if ( typeof(settings) == 'undefined') {
                _logger.error('Settings is not defined');
            }
            else if (typeof(settings.key) == 'undefined') {
                _logger.error('Settings.key is not defined');
            }
            else if ( typeof(settings.key) != 'string') {
                _logger.error('Settings.key is not a string (' + typeof(settings.key) + ')');
            }
            else if (settings.key.trim() == '') {
                _logger.error('Settings.key is empty string');
            }
            else {
                Config.API_KEY = settings.key;
                _logger.info('Config.API_KEY : ' + Config.API_KEY);
            }
        }
        catch(e) {
            _logger.error(e);
        }

        return Config.API_KEY;
    }

    /**
     * Save the API key to the settings file.
     * @param key
     * @returns {*}
     * @private
     */
    function _saveApiKey(key) {
        try {
            Config.API_KEY = key;
            return Utils.write_json_file(
                Config.SETTINGS_FILE,
                {key: key}
            );
        }
        catch(e) {
            return e;
        }
    }

    /**
     * Show the settings dialog.
     * @returns {string}
     * @private
     */
    function _showSettingsDialog() {
        var dialog;

        try {
            dialog = new SettingsDialog(Config);
            _logger.info(
                'Dialog is : {test}', {
                    test : typeof(dialog)
                });
            try {
                dialog.show();
            }
            catch(e) {
                _logger.error(new SettingsDialogError('dialog.show() failed : ' + $.stack))
            }
            return 'Dialog is a ' + typeof(dialog);
        }
        catch(e) {
            var errMessage = new SettingsDialogError('showSettingsDialog failed : ' + $.stack);
            _logger.error( errMessage );
            return errMessage;
        }
    }

    /**
     * Creates a web shortcut then opens it in the default browser.
     * @param address
     * @private
     */
    function _openInBrowser( address ) {
        try {
            Utils.write_exec(
                Folder.temp + '/' + now() + '-shortcut.url',
                '[InternetShortcut]' + '\r' + 'URL=' + encodeURI(address) + '\r'
            );
        }
        catch(e) {
            prompt("Visit " + address + " for more information.", address);
        }
    };

    /**
     * Create a new CSXSEvent.
     * @param   {string}    eventType
     * @param   {*}         data
     * @returns {CSXSEvent}
     * @private
     */
    function _getNewCSEvent(type, data) {

        _logger.info("Create new CSXSEvent(" + type + ", " + data + ")");

        var event = new CSXSEvent();
        event.type = type;
        event.data = data;
        return event;
    }

    /**
     * Load external object library.
     * @returns {boolean}
     * @private
     */
    function _loadExternalObject() {
        if (ExternalObjectIsLoadedLib) {
            _logger.info("[lib:PlugPlugExternalObject] already loaded");
            return true;
        }
        try {
            if (new ExternalObject("lib:\PlugPlugExternalObject")) {
                _logger.info("[lib:PlugPlugExternalObject] loaded", $.fileName, $.line);
                ExternalObjectIsLoadedLib = true;
                return ExternalObjectIsLoadedLib;
            }
        }
        catch (e) {
            _logger.error(e, $.fileName, $.line);
            throw new Error(e);
        }
    }

    /**
     * Confirm resizing the image before sending.
     * @returns {boolean}
     * @private
     */
    function _confirmImageResize() {
        return confirm(
            "Your image needs to be resized to meet our size limits. Would you like to resize and continue?"
        );
    }

    /**
     * Get size of selection or document.
     * @returns {{width: number, height: number}}
     * @private
     */
    function _getImageSize() {

        var selection = _getSelectionOrAll();

        _logger.info('getImageSize - selection : ' + typeof(selection));

        var width  = 0,
            height = 0;

        try {

            _logger.info('getImageSize - selection.bounds : ' + typeof(selection.bounds));

            width  = selection.bounds[2].as("px") - selection.bounds[0].as("px");
            height = selection.bounds[3].as("px") - selection.bounds[1].as("px");
        }
        catch(e) {
            _logger.error('getSelectionOrAll Error : ' + e);
        }

        _logger.info('typeof Math.toFixed : ' + typeof(Math.toFixed));

        _logger.info({
            width  : width,
            height : height,
            px     : width * height,
            mpix   : calcMegaPixels(width, height)
        }.toSource());

        return JSON.stringify({
            width  : width,
            height : height,
            px     : width * height,
            mpix   : calcMegaPixels(width, height)
        });
    }

    /**
     * Return the selected area or select all and return.
     * @returns {null|*}
     * @private
     */
    function _getSelectionOrAll() {

        var doc = activeDocument;

        _logger.info('getSelectionOrAll - doc : ' + doc.name);

        originalDocument = doc;
        currentSelection = doc.selection;
        hasSelection     = true;

        try {
            if (! currentSelection.bounds) {
                hasSelection = false;
            }
        }
        catch(e) {
            hasSelection = false;
        }

        _logger.info('getSelectionOrAll - hasSelection : ' + hasSelection);

        if (hasSelection) {
            originalSelection = [
                currentSelection.bounds[0].as("px"),
                currentSelection.bounds[1].as("px")
            ];
        }
        else {
            originalSelection = null;
            currentSelection.selectAll();
        }

        _logger.info('getSelectionOrAll - currentSelection : ' + typeof(currentSelection));

        return currentSelection;
    }

    /**
     * Get the name of the active doc.
     * @returns {*}
     */
    function _getActiveDocName() {
        return activeDocument.name;
    }

    /**
     * Create the USER_DATE folder if it does not exist.
     * @private
     */
    function _initDataFolder() {
        try {

            var dataFolder = Utils.folder(Config.USER_DATA);
            var logFolder  = Utils.folder(Config.LOG_FOLDER);

            return '_initDataFolder : ' + typeof(dataFolder) != 'undefined' && typeof(logFolder) != 'undefined';
        }
        catch(e) {
            alert('_initDataFolder : ' + e + ' - ' + $.fileName + ':' + $.line);
            return '_initDataFolder : ' + e;
        }
    }

    /**
     * Initialize the Host.
     * @param {string} extensionID
     */
    function _init() {

        var initFolders = _initDataFolder();

        if (typeof(Config.EXTENSION_ID) == 'undefined' && typeof($.global.extensionID) != 'undefined') {
            Config.EXTENSION_ID = $.global.extensionID;
        }
        $.global.logfolder = Config.LOG_FOLDER;
        $.global.Config    = Config;

        var apiKey = _getApiKey();
        if (typeof(apiKey) == 'string' && apiKey.length > 0) {
            Config.API_KEY = apiKey;
            $.global.API_KEY = apiKey;
        }
        _loadExternalObject();

        // app.notifiers.add("slct", new File("~/Downloads/test.js"), "Dcmn");

        // app.notifiers.add(
        //     "slct",
        //     File("~/Downloads/ps-notifier-test.js"),
        //     "Dcmn"
        // );

        return initFolders;
    }

    // Initialize the object.

    _init();

    /**
     * Return the public interface.
     */
    return {
        init : function() {
            return _init();
        },
        getApiKey : function() {
            return _getApiKey();
        },
        saveApiKey : function() {
            return _saveApiKey(key);
        },
        showSettingsDialog : function() {
            return _showSettingsDialog();
        },
        confirmImageResize : function() {
            return _confirmImageResize();
        },
        alert : function(message) {
            alert(message);
        },
        getImageSize : function() {
            return _getImageSize();
        },
        getActiveDocName : function() {
            return _getActiveDocName();
        },
        logger : _logger
    }
};

/**
 * Create an instance of the Host object.
 * @returns {boolean}
 */
function createHostInstance() {
    try {
        host = new Host(Config);
        return typeof(host) != 'undefined';
    }
    catch(e) {
        alert(e);
        return false;
    }
}

/**
 * Create a new custom event object.
 * @param type
 * @param data
 * @returns {CSXSEvent}
 * @constructor
 */
function CustomEvent(type, data) {
    var event  = new CSXSEvent();
    event.type = type;
    event.data = data;
    return event;
}

/**
 * Wrapper for CEP's evalFile.
 * @param theFilePath
 * @returns {*}
 */
function include(theFilePath) {
    try {
        if ((new File(theFilePath)).exists) {
            $.evalFile( theFilePath );
            return (new File(theFilePath)).name;
        }
        else {
            return theFilePath + ' does not exist';
        }
    }
    catch(e) {
        return e;
    }
}

/**
 * Escape JSON string.
 * @param str
 * @returns {void | string | never}
 */
function escapeJson(str) {
    return str.replace(/\\/g, "\\\\");
}

/**
 *
 * @param maxres
 * @returns {string}
 */
function verifyFileSize(maxres) {

    var doc,
        hasSelection,
        currentSelection,
        width,
        height,
        jpgOptions,
        file,
        megapix;

    doc = app.activeDocument;

    originalDocument = doc;
    currentSelection = doc.selection;
    hasSelection     = true;

    host.logger.info('------------------------------ VERIFY FILE SIZE (1) ------------------------------');

    try {
        if (! doc.selection.bounds) {
            hasSelection = false;
        }
        info(JSON.stringify({
            currentSelection : currentSelection.bounds,
            hasSelection : hasSelection
        }));
    }
    catch(e) {
        hasSelection = false;
    }

    if (hasSelection) {
        originalSelection = [
            currentSelection.bounds[0].as("px"),
            currentSelection.bounds[1].as("px")
        ];
    }
    else {
        originalSelection = null;
        doc.selection.selectAll();
    }

    host.logger.info('------------------------------ VERIFY FILE SIZE (2) ------------------------------');
    host.logger.info(JSON.stringify({
        originalSelection : originalSelection,
        hasSelection : hasSelection
    }));

    width  = doc.selection.bounds[2].as("px") - doc.selection.bounds[0].as("px");
    height = doc.selection.bounds[3].as("px") - doc.selection.bounds[1].as("px");

    megapix = (width * height) / 1000000;

    host.logger.info('------------------------------ VERIFY FILE SIZE (2) ------------------------------');
    host.logger.info(JSON.stringify({
        width  : width,
        height : height,
        size   : megapix
    }));

    // TODO: Remove Debug from production code.
    // maxres = megapix * 0.35;

    if ((width * height) > maxres) {

        host.logger.info('Image is too large. Will prompt user for resize confirmation');

        var confirmationMessage = _t(
            'Your image is {megapix} megapixels which exceeds the maximum size of {maxres} ' +
            'megapixels allowed by your plan. Would you like to resize the image then send?',
            { megapix : toFixed(megapix, 2), maxres : toFixed(maxres / 1000000, 2) }
        );

        host.logger.info(confirmationMessage);

        if ( ! confirm(confirmationMessage)) {
            host.logger.info('User declined to resize image');
            return '{"status": "cancel", "error": "user_cancelled"}';
        }

        host.logger.info('User opted to resize image');

        try {
            if (hasSelection) {

                host.logger.info('------------------------------ VERIFY FILE SIZE (3) ------------------------------');

                // Save the current selection in an alpha channel.

                selectionChannel = saveSelection();

                host.logger.info('Saved selection {s}', {s: selectionChannel.name});

                // Resize the image.

                resizeByMegapixels(maxres);

                host.logger.info('Resized image');

                // Re-load the selection from the alpha channel.
                loadSelection(selectionChannel);

                host.logger.info('Loaded previous selection {s}', {s: selectionChannel.name} );
            }
            else {
                host.logger.info('No previous selection. Using entire image');
                resizeByMegapixels(maxres);
                doc.selection.selectAll();
            }

            wasResized = true;

            host.logger.info(wasResized ? 'Image was resized' : 'Image was not resized');
        }
        catch(e) {
            host.logger.error(new ResizeImageError(e.message, e.stack));
            var err = new ResizeImageError('resize_image_error');
            return '{"status": "error", "error": "' + err.message + '"}';
        }
    }

    host.logger.info('------------------------------ VERIFY FILE SIZE (4) ------------------------------');

    host.logger.info('Attempting to copy selection');

    doc.selection.copy();

    width  = doc.selection.bounds[2].as("px") - doc.selection.bounds[0].as("px");
    height = doc.selection.bounds[3].as("px") - doc.selection.bounds[1].as("px");

    host.logger.info(JSON.stringify({
        width  : width,
        height : height,
        size   : (width * height) / 1000000
    }));

    host.logger.info('Creating temporary document');

    tempDoc = app.documents.add(
        new UnitValue(width, "px"),
        new UnitValue(height, "px"),
        72,
        "removebg",
        NewDocumentMode.RGB,
        DocumentFill.TRANSPARENT,
        1
    );

    tempDoc.paste();

    jpgOptions = new JPEGSaveOptions();

    jpgOptions.quality = 8;

    file = new File(File(Folder.temp).fsName + '/removebg.jpg');

    host.logger.info('Saving temporary document to {s}', {s: file.fsName});

    tempDoc.saveAs(file, jpgOptions, true, Extension.LOWERCASE);

    return '{"status": "ok", "path": "' + escapeJson(file.fsName) + '"}';
}

/**
 * Resize the current image to target megapixels.
 * @param width
 * @param height
 * @param maxres
 */
function resizeByMegapixels(maxres) {

    var startRulerUnits     = app.preferences.rulerUnits,
        startTypeUnits      = app.preferences.typeUnits,
        startDisplayDialogs = app.displayDialogs;

    var doc = app.activeDocument;

    try {
        app.displayDialogs         = DialogModes.NO;
        app.preferences.rulerUnits = Units.PIXELS;
        app.preferences.typeUnits  = TypeUnits.PIXELS;

        doc.changeMode(ChangeMode.RGB);

        var dWidth    = doc.width.value,
            dHeight   = doc.height.value,
            newWidth  = null,
            newHeight = null;

        host.logger.info('------------------------------ RESIZE BY MEGAPIXELS (1) ------------------------------');

        var myres = (dWidth * dHeight) / 1000000;
        var scale = maxres / (dWidth * dHeight);

        host.logger.info('Image resolution : {d}', {d: myres});
        host.logger.info('Resize scale : {d}', {d: scale});

        var _width  = Math.sqrt((dWidth * dWidth)   * scale),
            _height = Math.sqrt((dHeight * dHeight) * scale);

        host.logger.info('Document dimensions : W : {w}, H : {h}', {
            w : _width,
            h : _height
        });

        // calculateMegaPixels(doc.width, doc.height);

        if (dHeight > dWidth) {
            newHeight  = Math.sqrt((dHeight * dHeight) * scale);
            newHeight  = UnitValue(newHeight, "px");
        }
        else {
            newWidth  = Math.sqrt((dWidth * dWidth) * scale);
            newWidth  = UnitValue(newWidth, "px");
        }

        host.logger.info('New Document dimensions : W : {w}, H : {h}', {
            w : Math.sqrt((dWidth * dWidth) * scale),
            h : Math.sqrt((dHeight * dHeight) * scale)
        });

        host.logger.info(
            _t('Resize image to W : {nw}, H : {nh}, MP : {mp}', {
                nw : newWidth,
                nh : newHeight,
                mp : (_width * _height) / 1000000
            })
        );

        doc.resizeImage(newWidth, newHeight, null, ResampleMethod.BICUBIC);

        // Reset
        if (startDisplayDialogs !== undefined) { app.displayDialogs = startDisplayDialogs; }
        if (startRulerUnits     !== undefined) { app.preferences.rulerUnits = startRulerUnits; }
        if (startTypeUnits      !== undefined) { app.preferences.typeUnits = startTypeUnits; }

        return true;
    }
    catch(e) {

        host.logger.error('Resize By Megapixels error : {e}', {e : e});

        // Reset
        if (startDisplayDialogs !== undefined) { app.displayDialogs = startDisplayDialogs; }
        if (startRulerUnits     !== undefined) { app.preferences.rulerUnits = startRulerUnits; }
        if (startTypeUnits      !== undefined) { app.preferences.typeUnits = startTypeUnits; }

        throw new Error(e);
    }
}

/**
 * Generate a short unique ID with collision-checking.
 */
function shortUUID() {
    var _generatedUIDs = {};
    while (true) {
        var uid = ("0000" + ((Math.random() * Math.pow(36, 4)) | 0).toString(36)).slice(-4);
        if (!_generatedUIDs.hasOwnProperty(uid)) {
            _generatedUIDs[uid] = true;
            return uid;
        }
    }
}

/**
 * Save the current selection in an alpha channel mask.
 * @returns {Channel}
 */
function saveSelection() {
    var doc = activeDocument;

    var theSelection,
        theBounds,
        theChannel;

    try {
        theSelection = doc.selection;
        theSelection.contract("1px");
        theBounds    = theSelection.bounds;

        theChannel = doc.channels.add();
        theChannel.name = 'selection-' + shortUUID();

        doc.selection.store(theChannel);

        return theChannel;
    }
    catch(e) {
        return e;
    }
}

/**
 * Loads the passed Channel into a selection.
 * @param {Channel} theChannel	The Channel the selection was saved in.
 * @returns void
 */
function loadSelection(theChannel, doc) {

    if (typeof(doc) == 'undefined') {
        doc = activeDocument;
    }

    doc.selection.load(theChannel, SelectionType.REPLACE, false);

    doc.activeChannels = [
        app.activeDocument.channels.getByName('Red'),
        app.activeDocument.channels.getByName('Green'),
        app.activeDocument.channels.getByName('Blue'),
    ];

    return doc.selection;
}

/**
 * Cancel removing background.
 * @returns void
 */
function cancelRemoveBg() {
    tempDoc.close(SaveOptions.DONOTSAVECHANGES);
    originalDocument = null;
    tempDoc = null;
}

/**
 * Import the API result to layer from file.
 * @param   {string}    filename
 * @returns {string}
 */
function addLayerFromFile(filename) {

    var targetLayer = tempDoc.activeLayer;
    var file        = new File(filename);
    var newDoc      = open(file);
    var layer       = newDoc.activeLayer;
    var tempLayer   = layer.duplicate(tempDoc);

    try {
        newDoc.close(SaveOptions.DONOTSAVECHANGES);

        app.activeDocument = tempDoc;

        host.logger.info('------------------------------ ADD LAYER FROM FILE (1) ------------------------------');
        host.logger.info('Making layer mask from transparency');

        makeLayerMaskFromTransparency();

        host.logger.info('Unlock layer mask');

        unlockMask();

        host.logger.info('Move mask to layer');

        moveMaskToLayer(targetLayer.name);

        host.logger.info('Merge temp layer');

        tempLayer.merge();

        host.logger.info('Locking the layer mask');

        lockMask();

        var resultLayer = targetLayer.duplicate(originalDocument);

        tempDoc.close(SaveOptions.DONOTSAVECHANGES);

        app.activeDocument = originalDocument;
        originalDocument.selection.deselect();

        if (originalSelection != null) {

            host.logger.info('An original selection exists');

            if (wasResized && selectionChannel != null) {

                host.logger.info('The original image was resized, loadingn selection from alpha channel');

                originalDocument.selection.deselect();
                var tmpSelection = loadSelection(selectionChannel, originalDocument);
                var x = tmpSelection.bounds[0].as("px");
                var y = tmpSelection.bounds[1].as("px");
                originalDocument.selection.deselect();
                originalSelection = [x, y];
            }

            host.logger.info('Moving selection to original position');

            resultLayer.applyOffset(
                new UnitValue(originalSelection[0], "px"),
                new UnitValue(originalSelection[1], "px"),
                OffsetUndefinedAreas.WRAPAROUND
            );

            originalSelection = null;
        }

        host.logger.info('Hiding all layers except processed layer');

        for(var i = 1; i < originalDocument.artLayers.length; i++) {
            originalDocument.artLayers[i].visible = false;
        }

        originalDocument = null;
        tempDoc = null;

        app.runMenuItem(charIDToTypeID(("FtOn")));

        return "ok";
    }
    catch(e) {
        error('addLayerFromFile Error : ' + e);
        alert('addLayerFromFile Error : ' + e);
    }
}

function selectLayerPixels() {
    var id1268 = charIDToTypeID( "setd" );
    var desc307 = new ActionDescriptor();
    var id1269 = charIDToTypeID( "null" );
    var ref257 = new ActionReference();
    var id1270 = charIDToTypeID( "Chnl" );
    var id1271 = charIDToTypeID( "fsel" );
    ref257.putProperty( id1270, id1271 );
    desc307.putReference( id1269, ref257 );
    var id1272 = charIDToTypeID( "T   " );
    var ref258 = new ActionReference();
    var id1273 = charIDToTypeID( "Chnl" );
    var id1274 = charIDToTypeID( "Chnl" );
    var id1275 = charIDToTypeID( "Trsp" );
    ref258.putEnumerated( id1273, id1274, id1275 );
    desc307.putReference( id1272, ref258 );
    executeAction( id1268, desc307, DialogModes.NO )
}

function makeLayerMask(maskType) {
    if( maskType == undefined) maskType = 'RvlS' ; //from selection
    //requires a selection 'RvlS'  complete mask 'RvlA' otherThanSelection 'HdSl'
    var desc140 = new ActionDescriptor();
    desc140.putClass( charIDToTypeID('Nw  '), charIDToTypeID('Chnl') );
    var ref51 = new ActionReference();
    ref51.putEnumerated( charIDToTypeID('Chnl'), charIDToTypeID('Chnl'), charIDToTypeID('Msk ') );
    desc140.putReference( charIDToTypeID('At  '), ref51 );
    desc140.putEnumerated( charIDToTypeID('Usng'), charIDToTypeID('UsrM'), charIDToTypeID(maskType) );
    executeAction( charIDToTypeID('Mk  '), desc140, DialogModes.NO );
}

function makeLayerMaskFromTransparency() {
    var idMk = charIDToTypeID( "Mk  " );
    var desc22 = new ActionDescriptor();
    var idNw = charIDToTypeID( "Nw  " );
    var idChnl = charIDToTypeID( "Chnl" );
    desc22.putClass( idNw, idChnl );
    var idAt = charIDToTypeID( "At  " );
    var ref7 = new ActionReference();
    var idChnl = charIDToTypeID( "Chnl" );
    var idChnl = charIDToTypeID( "Chnl" );
    var idMsk = charIDToTypeID( "Msk " );
    ref7.putEnumerated( idChnl, idChnl, idMsk );
    desc22.putReference( idAt, ref7 );
    var idUsng = charIDToTypeID( "Usng" );
    var idUsrM = charIDToTypeID( "UsrM" );
    var idTrns = charIDToTypeID( "Trns" );
    desc22.putEnumerated( idUsng, idUsrM, idTrns );
    executeAction( idMk, desc22, DialogModes.NO );

}

function unlockMask() {
    var idsetd = charIDToTypeID( "setd" );
    var desc37 = new ActionDescriptor();
    var idnull = charIDToTypeID( "null" );
    var ref16 = new ActionReference();
    var idLyr = charIDToTypeID( "Lyr " );
    var idOrdn = charIDToTypeID( "Ordn" );
    var idTrgt = charIDToTypeID( "Trgt" );
    ref16.putEnumerated( idLyr, idOrdn, idTrgt );
    desc37.putReference( idnull, ref16 );
    var idT = charIDToTypeID( "T   " );
    var desc38 = new ActionDescriptor();
    var idUsrs = charIDToTypeID( "Usrs" );
    desc38.putBoolean( idUsrs, false );
    var idLyr = charIDToTypeID( "Lyr " );
    desc37.putObject( idT, idLyr, desc38 );
    executeAction( idsetd, desc37, DialogModes.NO );
}

function lockMask() {
    // =======================================================
    var idsetd = charIDToTypeID( "setd" );
    var desc27 = new ActionDescriptor();
    var idnull = charIDToTypeID( "null" );
    var ref10 = new ActionReference();
    var idLyr = charIDToTypeID( "Lyr " );
    var idOrdn = charIDToTypeID( "Ordn" );
    var idTrgt = charIDToTypeID( "Trgt" );
    ref10.putEnumerated( idLyr, idOrdn, idTrgt );
    desc27.putReference( idnull, ref10 );
    var idT = charIDToTypeID( "T   " );
    var desc28 = new ActionDescriptor();
    var idUsrs = charIDToTypeID( "Usrs" );
    desc28.putBoolean( idUsrs, true );
    var idLyr = charIDToTypeID( "Lyr " );
    desc27.putObject( idT, idLyr, desc28 );
    executeAction( idsetd, desc27, DialogModes.NO );
}

function moveMaskToLayer(layerName) {
    // =======================================================
    var idslct = charIDToTypeID( "slct" );
    var desc39 = new ActionDescriptor();
    var idnull = charIDToTypeID( "null" );
    var ref17 = new ActionReference();
    var idChnl = charIDToTypeID( "Chnl" );
    var idChnl = charIDToTypeID( "Chnl" );
    var idMsk = charIDToTypeID( "Msk " );
    ref17.putEnumerated( idChnl, idChnl, idMsk );
    desc39.putReference( idnull, ref17 );
    var idMkVs = charIDToTypeID( "MkVs" );
    desc39.putBoolean( idMkVs, false );
    executeAction( idslct, desc39, DialogModes.NO );

    // =======================================================
    var idMk = charIDToTypeID( "Mk  " );
    var desc40 = new ActionDescriptor();
    var idNw = charIDToTypeID( "Nw  " );
    var idChnl = charIDToTypeID( "Chnl" );
    desc40.putClass( idNw, idChnl );
    var idAt = charIDToTypeID( "At  " );
    var ref18 = new ActionReference();
    var idChnl = charIDToTypeID( "Chnl" );
    var idChnl = charIDToTypeID( "Chnl" );
    var idMsk = charIDToTypeID( "Msk " );
    ref18.putEnumerated( idChnl, idChnl, idMsk );
    var idLyr = charIDToTypeID( "Lyr " );
    ref18.putName( idLyr, layerName );
    desc40.putReference( idAt, ref18 );
    var idUsng = charIDToTypeID( "Usng" );
    var ref19 = new ActionReference();
    var idChnl = charIDToTypeID( "Chnl" );
    var idChnl = charIDToTypeID( "Chnl" );
    var idMsk = charIDToTypeID( "Msk " );
    ref19.putEnumerated( idChnl, idChnl, idMsk );
    var idLyr = charIDToTypeID( "Lyr " );
    var idOrdn = charIDToTypeID( "Ordn" );
    var idTrgt = charIDToTypeID( "Trgt" );
    ref19.putEnumerated( idLyr, idOrdn, idTrgt );
    desc40.putReference( idUsng, ref19 );
    executeAction( idMk, desc40, DialogModes.NO );
}
